package cn.com.qmth.examcloud.core.oe.task.service.worker;

import cn.com.qmth.examcloud.commons.exception.StatusException;
import cn.com.qmth.examcloud.commons.helpers.concurrency.simple.Worker;
import cn.com.qmth.examcloud.commons.helpers.concurrency.simple.WorkerController;
import cn.com.qmth.examcloud.commons.util.Util;
import cn.com.qmth.examcloud.core.oe.task.base.ExamCaptureProcessStatisticController;
import cn.com.qmth.examcloud.core.oe.task.dao.enums.ExamCaptureQueueStatus;
import cn.com.qmth.examcloud.core.oe.task.service.ExamCaptureQueueService;
import cn.com.qmth.examcloud.core.oe.task.service.ExamCaptureService;
import cn.com.qmth.examcloud.core.oe.task.service.bean.ExamCaptureQueueInfo;
import cn.com.qmth.examcloud.support.Constants;
import cn.com.qmth.examcloud.web.support.SpringContextHolder;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.logging.log4j.ThreadContext;

import java.util.concurrent.TimeUnit;

/**
 * 百度活体检测工作线程
 *
 * @author lideyin 20190620
 */
public class BaiduFaceLivenessWorker implements Worker<ExamCaptureQueueInfo> {

    private final Log captureLog = LogFactory.getLog("PROCESS_EXAM_CAPTURE_TASK_LOGGER");

    @Override
    public void process(WorkerController controller, ExamCaptureQueueInfo element) {
        ThreadContext.put("TRACE_ID", "Q_" + element.getId());
        ThreadContext.put("CALLER", "BAIDU_WORKER");
        ExamCaptureProcessStatisticController.increaseFaceLivenessDetectCount();
        if (captureLog.isDebugEnabled()) {
            captureLog.debug("[BAIDU_FACELIVENESS_WORKER.] 图片处理数量+1，count= "
                    + ExamCaptureProcessStatisticController.getFaceLivenessDetectCount());
        }
        ExamCaptureService examCaptureService = SpringContextHolder
                .getBean(ExamCaptureService.class);
        ExamCaptureQueueService examCaptureQueueService = SpringContextHolder
                .getBean(ExamCaptureQueueService.class);
        try {
            examCaptureService.disposeBaiDuFaceLiveness(element);
        } catch (StatusException e) {
            while (true) {
                // 异常处理
                if (examCaptureQueueService.saveExamCaptureQueueEntityByFailed(element.getId(), e.getDesc(),
                        ExamCaptureQueueStatus.PROCESS_FACELIVENESS_FAILED)) {
                    break;
                }
                ;
                Util.sleep(TimeUnit.MILLISECONDS, 500);
            }
            if ((e.getCode().equals(Constants.BAIDU_FACELIVENESS_QPS_LIMIT_EXCEEDED_CODE))) {
                // 如果超过并发次数，则添加异常次数
                controller.addConcurrencyWarn();
                captureLog.debug("[BAIDU_FACELIVENESS_WORKER.] 超过并发次数 ");
            } else {
                ExamCaptureProcessStatisticController.increaseFaceLivenessDetectCount();
                captureLog.error("[BAIDU_FACELIVENESS_WORKER.] 自定义异常 " + e.getDesc(), e);
            }
            Util.sleep(TimeUnit.MICROSECONDS, 50);
        } catch (Exception e) {
            ExamCaptureProcessStatisticController.increaseFaceLivenessDetectCount();
            while (true) {
                // 异常处理
                if (examCaptureQueueService.saveExamCaptureQueueEntityByFailed(element.getId(), "系统异常",
                        ExamCaptureQueueStatus.PROCESS_FACELIVENESS_FAILED)) {
                    break;
                }
                ;
                Util.sleep(TimeUnit.MILLISECONDS, 500);
            }
            captureLog.error("[BAIDU_FACELIVENESS_WORKER.] 系统异常 " + e.getMessage(), e);
            Util.sleep(TimeUnit.MICROSECONDS, 50);
        } finally {
            ThreadContext.clearAll();
        }
    }

}
